/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "pic32_c.h"
#include "init.h"
#include "vga_display.h"



/* Display state. */

vga_display_t vga_display;

/* Pixel data. */

static const int ZERO_LENGTH = 1;
static const uint8_t zerodata[1] = {0};



/* Initialise the state machine. */

void init_vga(display_config_t *display_config, int line_channels,
              int initiating_int_num)
{
    /* Display parameters. */

    vga_display.display_config = display_config;
    vga_display.line_channels = line_channels;
    vga_display.initiating_int_num = initiating_int_num;

    /* Initial state. */

    vga_display.state_handler = vbp_active;
    vga_display.line = 0;
}



/* Configure DMA channels for the transfer of pixel data. */

void vga_configure_transfer(int transfer_int_num, uint32_t output)
{
    int initiating_channel = vga_display.initiating_int_num >= 0;
    int dual_channel = vga_display.line_channels == 2;
    int channel = 0;

    /* Where dual line channels are involved, put the first before any
       initiating channel, chaining it to such a channel. */

    if (dual_channel)
    {
        vga_configure_line_channel(channel++, transfer_int_num,
            initiating_channel ? dma_chain_next : dma_chain_none,
            output);
    }

    /* Introduce a special initiating channel if an interrupt has been
       indicated. */

    if (initiating_channel)
    {
        vga_configure_zero_channel(channel++, vga_display.initiating_int_num, 1,
                                   output);
    }

    /* A line channel is always configured, chaining it to any initiating
       channel. */

    vga_configure_line_channel(channel++, transfer_int_num,
        initiating_channel ? dma_chain_previous : dma_chain_none,
        output);

    /* A zero channel is always configured, chaining it to the preceding line
       channel, with the same transfer interrupt initiating the transfer. */

    vga_configure_zero_channel(channel, transfer_int_num, 0, output);
}

void vga_configure_line_channel(int channel, int int_num, enum dma_chain chain,
                                uint32_t output)
{
    dma_init(channel, 2);

    /* If initiating, configure to always be enabled. */

    if (chain == dma_chain_none)
        dma_set_auto_enable(channel, 1);

    /* Chain to either the previous or next channel where a special initiating
       channel is used. */

    else
        dma_set_chaining(channel, chain);

    /* Initiate DMA on the transfer interrupt, transferring line data to the
       first byte of the output word. */

    dma_set_interrupt(channel, int_num, 1);

    dma_set_destination(channel, HW_PHYSICAL(output), 1);
    dma_set_cell(channel, vga_display.display_config->transfer_cell_size);
}

void vga_configure_zero_channel(int channel, int int_num, int initiating,
                                uint32_t output)
{
    dma_init(channel, 3);

    /* If initiating, configure to always be enabled. */

    if (initiating)
        dma_set_auto_enable(channel, 1);

    /* Otherwise, chain to the preceding channel and register prior transfer
       events before the channel is activated. */

    else
    {
        dma_set_chaining(channel, dma_chain_previous);
        dma_set_receive_events(channel, 1);
    }

    /* Transfer upon the indicated interrupt condition. */

    dma_set_interrupt(channel, int_num, 1);
    dma_set_transfer(channel, PHYSICAL((uint32_t) zerodata),
                              ZERO_LENGTH,
                              HW_PHYSICAL(output), 1,
                              ZERO_LENGTH);
}

/* Configure a timer and output compare units for horizontal and vertical
   sync. */

void vga_configure_sync(int hsync_unit, int vsync_unit, int timer)
{
    /* Record the peripherals in use. */

    vga_display.hsync_unit = hsync_unit;
    vga_display.vsync_unit = vsync_unit;
    vga_display.timer = timer;

    /* Configure a timer for the horizontal sync. The timer has no prescaling
       (0). */
    
    timer_init(timer, 0, vga_display.display_config->hfreq_limit);
    timer_on(timer);
    
    /* Horizontal sync. */
    
    /* Configure output compare in dual compare (continuous output) mode using
       the timer as time base. The interrupt condition drives the first DMA
       channel and is handled to drive the display state machine. */

    oc_init(hsync_unit, 0b101, timer);
    oc_set_pulse(hsync_unit, vga_display.display_config->hsync_end);
    oc_set_pulse_end(hsync_unit, vga_display.display_config->hsync_start);
    oc_init_interrupt(hsync_unit, 7, 3);
    oc_on(hsync_unit);

    /* Vertical sync. */

    /* Configure output compare in single compare (output driven low) mode using
       the timer as time base. The unit is enabled later. It is only really used
       to achieve precisely-timed level transitions in hardware. */
                        
    oc_init(vsync_unit, 0b010, timer);
    oc_set_pulse(vsync_unit, 0);
}



/* Interrupt handlers. */

void vga_interrupt_handler(void)
{
    vga_display.line += 1;
    vga_display.state_handler();
}



/* Vertical back porch region. */

void vbp_active(void)
{
    if (vga_display.line < vga_display.display_config->visible_start)
        return;

    /* Enter the visible region. */

    vga_display.state_handler = visible_active;

    /* Set the line address. */

    vga_display.linedata = vga_display.display_config->screen_start;
    start_visible();
}

/* Visible region. */

void visible_active(void)
{
    if (vga_display.line < vga_display.display_config->vfp_start)
    {
        /* Update the line address and handle wraparound. */

        if (!(vga_display.line % vga_display.display_config->line_multiplier))
        {
            vga_display.linedata += vga_display.display_config->line_length;

            if (vga_display.linedata >= vga_display.display_config->screen_limit)
                vga_display.linedata -= vga_display.display_config->screen_size;
        }

        update_visible();
        return;
    }

    /* End the visible region. */

    vga_display.state_handler = vfp_active;

    /* Disable the channel for the next line. */

    stop_visible();
}

/* Vertical front porch region. */

void vfp_active(void)
{
    if (vga_display.line < vga_display.display_config->vsync_start)
        return;

    /* Enter the vertical sync region. */

    vga_display.state_handler = vsync_active;

    /* Bring vsync low when the next line starts. */

    vsync_low();
}

/* Vertical sync region. */

void vsync_active(void)
{
    if (vga_display.line < vga_display.display_config->vsync_end)
        return;

    /* Start again at the top of the display. */

    vga_display.line = 0;
    vga_display.state_handler = vbp_active;

    /* Bring vsync high when the next line starts. */

    vsync_high();
}



/* Enable the channels for the next line. */

void start_visible(void)
{
    update_visible();
    update_transfers(1);
}

/* Update the channels for the next line. */

void update_visible(void)
{
    uint32_t transfer_start = (uint32_t) vga_display.linedata;
    uint32_t transfer_length = vga_display.display_config->line_length /
                               vga_display.line_channels;

    /* Determine whether an initiating channel is used. */

    int initiating_channel = vga_display.initiating_int_num >= 0;
    int channel = 0;

    /* Update the source of a secondary line channel. */

    if (vga_display.line_channels == 2)
    {
        dma_set_source(channel, PHYSICAL(transfer_start), transfer_length);
        transfer_start += transfer_length;
        channel++;
    }

    /* Skip any initiating channel. */

    if (initiating_channel)
        channel++;

    /* Update the source of a line channel, with potentially updated start
       address. */

    dma_set_source(channel, PHYSICAL(transfer_start), transfer_length);
}

/* Disable the channels for the next line. */

void stop_visible(void)
{
    update_transfers(0);
}

/* Enable or disable transfers. */

void update_transfers(int enable)
{
    int initiating_channel = vga_display.initiating_int_num >= 0;
    int channel = 0;
    void (*fn)() = enable ? dma_on : dma_off;

    /* Update line channels if no initiating channel is used. */

    if (vga_display.line_channels == 2)
    {
        if (!initiating_channel)
            fn(channel);
        channel++;
    }

    /* Update any initiating channel. */

    if (initiating_channel)
    {
        fn(channel);
        channel++;
    }

    /* Update line channels if no initiating channel is used. */

    if (!initiating_channel)
        fn(channel);
}



/* Bring vsync low (single compare, output driven low) when the next line
   starts. */

void vsync_low(void)
{
    oc_init(vga_display.vsync_unit, 0b010, vga_display.timer);
    oc_on(vga_display.vsync_unit);
}

/* Bring vsync high (single compare, output driven high) when the next line
   starts. */

void vsync_high(void)
{
    oc_init(vga_display.vsync_unit, 0b001, vga_display.timer);
    oc_on(vga_display.vsync_unit);
}
